@@ -103,7 +103,8 @@ on_window1_realize                     (GtkWidget       *widget,
                                         gpointer         user_data)
 {
   gboolean is_dock = (gboolean) user_data;
-  
+  GdkKeymap *map = gdk_keymap_get_default ();
+
   if (is_dock) {
 	  gok_main_set_wm_dock (TRUE);
   }
@@ -125,6 +126,9 @@ on_window1_realize                     (GtkWidget       *widget,
   gdk_window_add_filter (NULL,
 			 gok_xkb_filter,
 			 NULL);
+
+  g_signal_connect (map, "keys-changed",
+		    (GCallback) gok_spy_keymap_listener, NULL);
 }
 
 
