@@ -535,7 +535,7 @@ void gok_editor_add_key ()
 		pNewKey->Left = col;
 		pNewKey->Right = col + 1;
 		gok_log("adding key {top [%d] bottom [%d] left [%d] right [%d]}",pNewKey->Top,pNewKey->Bottom,pNewKey->Left,pNewKey->Right);
-		gok_key_add_label (pNewKey, _("label"), 0, 0);
+		gok_key_add_label (pNewKey, _("label"), 0, 0, NULL);
 		
 		gok_editor_touch_file(TRUE);
 		gok_editor_update_keyboard(m_pKeyboard);
@@ -808,10 +808,9 @@ gboolean gok_editor_save_keyboard (GokKeyboard* pKeyboard, gchar* Filename)
 		while (pLabel != NULL)
 		{
 			fputs ("\t\t<GOK:label", pFile);
-			if (pLabel->modifiers != 0)
-			{
-				fprintf (pFile, " modifier=\"%s\"", 
-					 gok_modifier_first_name_from_mask (pLabel->modifiers));
+			if (pLabel->level != 0)			{
+				fprintf (pFile, " level=\"%d\"", 
+					 pLabel->level);
 			}
 			fprintf (pFile, ">%s</GOK:label>\n", pLabel->Text);
 			
