@@ -50,16 +50,45 @@ long keysym2ucs(KeySym keysym); /* from keysym2ucs.c */
 
 gchar m_TextEmpty[]="";
 
-char m_label[XkbKeyNameLength + 1];
+gchar m_label[XkbKeyNameLength + 1];
 
 gint m_Group = 0; /* FIXME: statics are bad */
 
+#undef GOK_UTF8_DEBUG
+
+static void 
+gok_key_print_utf8_debug (gchar *buf)
+{
+      gchar *cp = buf;
+      gunichar next_char;
+      while ((next_char = g_utf8_get_char_validated (cp, -1)) >= 0) 
+      {
+	  fprintf (stderr, "[%x]", (unsigned long) next_char);
+	  if (next_char == 0) 
+	  {
+	      fprintf (stderr, "\t");
+	      break;
+	  }
+	  cp = g_utf8_next_char (cp);
+      }
+      cp = buf;
+      while (*cp)
+      {
+	  guchar c = *cp;
+	  fprintf (stderr, "0x%x ", (unsigned) c);
+	  ++cp;
+      }
+      fprintf (stderr, "\n");
+}
+
 /* returned string must be freed */
-static char *
-gok_key_lookup_shorter_label (char *str)
+static guchar *
+gok_key_lookup_shorter_label (guchar *str)
 {
 	gint i;
-	gchar *xstr = NULL;
+	guchar *xstr = NULL;
+
+	/* handle deadkeys gracefully */
 
 	if (!strcmp (str, "Delete")) {
 		xstr = "Del";
@@ -94,6 +123,66 @@ gok_key_lookup_shorter_label (char *str)
 	else if (!strcmp (str, "Sterling")) {
 		xstr = "£";
 	}
+	else if (!strncmp (str, "dead\n", 5)) {
+		gchar *cp = (gchar *) &str[5];
+		if (!strcmp (cp, "grave")) {
+			xstr = "[\140]";
+		}
+		else if (!strcmp (cp, "acute")) {
+			xstr = "[\302\264]";
+		}
+		else if (!strcmp (cp, "circumflex")) {
+			xstr = "[^]";
+		}
+		else if (!strcmp (cp, "tilde")) {
+			xstr = "[~]";
+		}
+		else if (!strcmp (cp, "macron")) {
+			xstr = "[\302\257]";
+		}
+		else if (!strcmp (cp, "breve")) {
+			xstr = "[\313\230]";
+		}
+		else if (!strcmp (cp, "abovedot")) {
+			xstr = "[\313\231]";
+		}
+		else if (!strcmp (cp, "diaeresis")) {
+			xstr = "[\302\250]";
+		}
+		else if (!strcmp (cp, "abovering")) {
+			xstr = "[\313\232]";
+		}
+		else if (!strcmp (cp, "doubleacute")) {
+			xstr = "[\313\235]";
+		}
+		else if (!strcmp (cp, "caron")) {
+			xstr = "[\313\207]";
+		}
+		else if (!strcmp (cp, "cedilla")) {
+			xstr = "[\302\270]";
+		}
+		else if (!strcmp (cp, "ogonek")) {
+			xstr = "[\313\233]";
+		}
+		else if (!strcmp (cp, "iota")) {
+			xstr = "[\315\205]";
+		}
+		else if (!strcmp (cp, "voiced")) { /* not sure */
+			xstr = "[\313\254]";
+		}
+		else if (!strncmp (cp, "semivoiced", 10)) { /* not sure */
+			xstr = "[\312\261]";
+		}
+		else if (!strcmp (cp, "belowdot")) {
+			xstr = "[\314\243]";
+		}
+		else if (!strcmp (cp, "hook")) {
+			xstr = "[\313\236]";
+		}
+		else if (!strcmp (cp, "horn")) { /* not sure */
+			xstr = "[\314\233]";
+		}
+	}
 
 	if (xstr != NULL) {
 		g_free (str);
@@ -102,7 +191,7 @@ gok_key_lookup_shorter_label (char *str)
 	else {
 		gint len;
 		gboolean prev_is_lower, is_upper;
-		gchar *s1, *st2, *st1;
+		guchar *s1, *st2, *st1;
 		
 		if (!strncmp (str, "Sun", 3)) {
 			s1 = g_strdup (str + 3);
@@ -130,13 +219,15 @@ gok_key_lookup_shorter_label (char *str)
 	return str;
 }
 
-static char *	
-gok_key_label_from_keysym_string (const char *str)
+static guchar *	
+gok_key_label_from_keysym_string (const guchar *str)
 {
-	char *cp;
+	guchar *cp;
 	size_t strl;
-	if (!str) return "<nil>";
+
+	if (!str) return g_strdup ("<nil>");
 	strl = strlen(str);
+
 	if (strncmp(str, "KP_", 3) == 0) {
 	        cp = g_strdup (str + 3);
 	}
@@ -149,11 +240,12 @@ gok_key_label_from_keysym_string (const char *str)
 	else {
 	        cp = g_strdup (str);
 	}
-	if (g_strrstr (cp, "_")) {
+	if ((strl > 1) && g_strrstr (cp, "_")) {
 		*g_strrstr (cp, "_") = '\n';
 	}
 	if (strlen (cp) > 4)
 		cp = gok_key_lookup_shorter_label (cp);
+
 	return cp;
 }
 
@@ -171,8 +263,8 @@ gok_key_get_numlock_mask (Display *display)
 }
 
 
-static int
-gok_key_get_xkb_type_index (XkbDescPtr xkb, KeyCode keycode, int group)
+int
+gok_key_get_xkb_type_index (XkbDescPtr xkb, KeyCode keycode, guint group)
 {
 	int num_groups = XkbKeyNumGroups (xkb, keycode);
 	int index;
@@ -183,35 +275,69 @@ gok_key_get_xkb_type_index (XkbDescPtr xkb, KeyCode keycode, int group)
 	return xkb->map->key_sym_map[keycode].kt_index[group];
 }
 
-static int
+int
 gok_key_level_for_type (Display *display, XkbDescRec *kbd,
 			int type, unsigned int *modmask)
 {
 	int i, level = 0;
+	gboolean level_set = FALSE;
 	XkbKeyTypeRec *key_type = &kbd->map->types[type];
+
 	/* compare against each map entry in the XkbKeyType */
-	for (i = 0; i < key_type->map_count; ++i)
+	for (i = 0; i < key_type->map_count && !level_set; ++i)
 	{
-		unsigned int vmods_equiv = 0, level_mask = 0;
+		unsigned int vmods_equiv = 0, level_mods = 0;
 		unsigned int mods_mask;
+		unsigned int mods = *modmask;
 		XkbVirtualModsToReal (kbd, key_type->map[i].mods.vmods,
 				      &vmods_equiv);
 		XkbVirtualModsToReal (kbd, key_type->mods.vmods,
 				      &mods_mask);
-		mods_mask = key_type->mods.mask | mods_mask;
+		mods_mask = key_type->mods.mask;
+		level_mods = key_type->map[i].mods.real_mods | vmods_equiv;
 #ifdef GOK_DEBUG
-		if (type == 3) fprintf (stderr, "keypad [%d], level %d, realmods = %x, vmods = %x, preserve %x; mods_mask %x\n",
-					i, key_type->map[i].level,
-					key_type->map[i].mods.real_mods,
-					vmods_equiv,
+		fprintf (stderr, "key type %d [map %d], level %d, preserve %x; mods_mask %x; level_mods %x; modmask %x\n",
+					type, i, key_type->map[i].level,
 					key_type->preserve ? key_type->preserve[i].mask : 0,
-					mods_mask);
+					mods_mask,
+			                level_mods,
+			                *modmask);
 #endif
-		level_mask = key_type->map[i].mods.mask | vmods_equiv;
 		if ((key_type->map[i].active) && 
-		    ((*modmask & mods_mask) == level_mask))
+		    ((mods & mods_mask) == level_mods))
 		{	
 			level = key_type->map[i].level;
+			level_set = TRUE;
+			/** 
+			 *  Special hack for alpha types: see section 15.2.1 of XKBlib spec. 
+			 *  We'd restrict this to the canonical types, but XFree's maps tend to use 
+			 *  non-canonical types for the primary alphanumeric keycodes, in order
+			 *  to support more shift levels (AltGr, Shift-AltGr, etc.)
+			 *  If we reported the Xkb 'level' here, the keycap strings would not get
+			 *  'shifted' in our compose keyboard when CapsLock is active, since the 
+			 *  capitalization in that case is deferred to the Xlib lookup stage.
+			 **/
+			if (key_type->preserve && (key_type->preserve[i].mask == LockMask))
+			{
+			    /* Lock is preserved for Xlib, but we want to report 'equivalent' level */
+			    if ((mods & LockMask) && !(mods & ShiftMask))
+			    {
+				int j;
+				/* find the map matching the 'Shift' equivalent */
+				guint equiv_mods = (mods & ~LockMask) | ShiftMask; 
+				for (j = 0; j < key_type->map_count; ++j)
+				{
+				    if (key_type->map[j].active && 
+					(key_type->map[j].mods.real_mods == equiv_mods)) 
+				    {
+					level = key_type->map[j].level;
+					level_set = TRUE;
+					break;
+				    }
+				}
+			    }
+			}
+
 			/* preserve what needs preserving */
 			if (key_type->preserve) 
 				*modmask = *modmask & key_type->preserve[i].mask;
@@ -226,41 +352,31 @@ gok_key_level_for_type (Display *display, XkbDescRec *kbd,
 	return level;
 }
 
-static char *
-gok_key_label_from_keycode (KeyCode keycode, Display *display, unsigned int modmask, gint group)
+static gchar *
+gok_key_label_from_keycode (KeyCode keycode, Display *display, guint level, guint group)
 {
   char buf[20];
   int extra_rtn, nbytes, state = 0;
   KeySym keysym;
   XkbStateRec xkb_state;
-  int level;
-  int type;
   long ucs;
+  gunichar unichar;
 
-  if (modmask == GOK_MODMASK_CURRENT) {
-          XkbGetState (display, XkbUseCoreKbd, &xkb_state);
-          modmask = xkb_state.compat_state;
-  }
-  type =  gok_key_get_xkb_type_index (gok_keyboard_get_xkb_desc (), keycode, group);
-  level = gok_key_level_for_type (display, gok_keyboard_get_xkb_desc(), type, &modmask);
   keysym = XkbKeycodeToKeysym (display, keycode, group, level);
   if ((ucs = keysym2ucs (keysym)) <= 0) 
   {
       /* workaround; XSun gets nbytes wrong, so we truncate to the largest legal value */  
-      nbytes = MIN (XkbTranslateKeySym (display, &keysym, modmask, buf, 19, &extra_rtn), 19); 
-      if (nbytes && g_unichar_isgraph (g_utf8_get_char_validated (buf, nbytes))) {
-	  return g_strndup (buf, nbytes);
-      }
-      else {  /* try again, without modifiers...*/
-	  nbytes = MIN (XkbTranslateKeySym (display, &keysym, 0, buf, 19, &extra_rtn), 19);
-	  return gok_key_label_from_keysym_string (
-	      (nbytes && g_unichar_isgraph (g_utf8_get_char_validated (buf, nbytes))) ? buf : XKeysymToString (keysym));
-      }
+      nbytes = MIN (XkbTranslateKeySym (display, &keysym, 0, buf, 19, &extra_rtn), 19);
+      buf[nbytes] = '\0';
+      return gok_key_label_from_keysym_string (XKeysymToString (keysym));
   }
   else
   {
-      buf[g_unichar_to_utf8 (ucs, buf)] = '\0';
-      return gok_key_label_from_keysym_string (buf);
+      gchar cbuf[10];
+      unichar = (gunichar) ucs;
+      cbuf[g_unichar_to_utf8 (unichar, cbuf)] = '\0';
+
+      return gok_key_label_from_keysym_string (cbuf);
   }
 }
 
@@ -268,7 +384,7 @@ static int
 gok_key_keycode_from_xkb_key (XkbKeyPtr keyp, XkbDescPtr kbd)
 {
   int k;
-  char *name = keyp->name.name;
+  gchar *name = keyp->name.name;
   if (kbd) {
     for (k = kbd->min_key_code; k < kbd->max_key_code; ++k) 
       {
@@ -317,7 +433,7 @@ gok_key_get_default_border_height (GokKey *pKey)
 gchar *
 gok_key_modifier_for_keycode (Display *display, XkbDescPtr xkb, int keycode)
 {
-	gint group = gok_key_get_effective_group ();
+	guint group = gok_key_get_effective_group ();
 	gint type = gok_key_get_xkb_type_index (xkb, keycode, group);
 	guint modmask = 0;	
 	GokModifier *modifier;
@@ -375,21 +491,17 @@ GokKey *
 gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display, XkbGeometryPtr pGeom, XkbRowPtr pRow, XkbSectionPtr pXkbSection, XkbKeyPtr keyp, int section, int row, int col)
 {
   GokKey *pKey;
-  char *label;
-  char name[XkbKeyNameLength + 1];
-  char keycode_name[8];
+  gchar *label;
+  gchar name[XkbKeyNameLength + 1];
+  gchar keycode_name[8];
   int keycode;
   XkbBoundsRec *pBounds;
   int width, height, len;
   gboolean is_modifier = FALSE;
   unsigned int numlock_mask;
-  guint modmask = SPI_KEYMASK_SHIFT | SPI_KEYMASK_SHIFTLOCK | SPI_KEYMASK_ALT | 
-	  SPI_KEYMASK_CONTROL | SPI_KEYMASK_MOD1 | SPI_KEYMASK_MOD2 | SPI_KEYMASK_MOD3 | SPI_KEYMASK_MOD4 |
-	  SPI_KEYMASK_MOD5; 
+  int group, level;
 
   pKey = gok_key_new (prevKey, NULL, pKeyboard);
-  numlock_mask = gok_key_get_numlock_mask (display);
-  modmask |= numlock_mask;
   is_modifier = FALSE;
   pKey->has_text = TRUE; /* TODO: handle non-textual keyboard key caps! */
   pKey->Top = row;
@@ -406,7 +518,7 @@ gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display,
 	  is_modifier = TRUE;
 	  gok_modifier_add (pKey->ModifierName);
   }
-  if (is_modifier) { /* FIXME */
+  if (is_modifier) {
     pKey->Type = KEYTYPE_MODIFIER;
     pKey->pImage = gok_keyimage_new (pKey, NULL);
     pKey->pImage->type = IMAGE_TYPE_INDICATOR;
@@ -416,55 +528,33 @@ gok_key_from_xkb_key (GokKey *prevKey, GokKeyboard *pKeyboard, Display *display,
 	  pKey->is_repeatable = TRUE;
   }
   pKey->Style = KEYSTYLE_NORMAL;
-  label = gok_key_label_from_keycode (keycode, display, 0, 0);
+
   pKey->FontSize = -1;
-  len = g_utf8_strlen (label, -1);
-  if (len > 1) {
-          if (len < 4) 
-                  pKey->FontSizeGroup = 3;
-	  else
-	          pKey->FontSizeGroup = FONT_SIZE_GROUP_UNIQUE;  
-  }
-  else {
-          pKey->FontSizeGroup = FONT_SIZE_GROUP_GLYPH;
-  }
+
   /* TODO: figure out how to specify smaller modmasks */
-  gok_keylabel_new (pKey, label, 0, modmask);
-  label = gok_key_label_from_keycode (keycode, display, ShiftMask, 0);
-  gok_keylabel_new (pKey, label, ShiftMask, modmask);
-  label = gok_key_label_from_keycode (keycode, display, LockMask | ShiftMask, 0);
-  gok_keylabel_new (pKey, label, LockMask | ShiftMask, modmask);
-  label = gok_key_label_from_keycode (keycode, display, LockMask, 0);
-  gok_keylabel_new (pKey, label, LockMask, modmask);
-  label = gok_key_label_from_keycode (keycode, display, Mod1Mask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_ALT, modmask);
-  label = gok_key_label_from_keycode (keycode, display, ControlMask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_CONTROL, modmask);
-  label = gok_key_label_from_keycode (keycode, display, numlock_mask, 0);
-  gok_keylabel_new (pKey, label, numlock_mask, modmask);
-  label = gok_key_label_from_keycode (keycode, display, numlock_mask | ShiftMask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_SHIFT | numlock_mask, 
-		    modmask);
-  label = gok_key_label_from_keycode (keycode, display, numlock_mask | LockMask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_SHIFTLOCK | numlock_mask, 
-		    modmask);
-  label = gok_key_label_from_keycode (keycode, display, 
-				      numlock_mask | LockMask | ShiftMask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_SHIFTLOCK | SPI_KEYMASK_SHIFT | numlock_mask, 
-		    modmask);
-  label = gok_key_label_from_keycode (keycode, display, Mod3Mask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD2, modmask);
-  label = gok_key_label_from_keycode (keycode, display, Mod3Mask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD3, modmask);
-  label = gok_key_label_from_keycode (keycode, display, Mod4Mask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD4, modmask);
-  label = gok_key_label_from_keycode (keycode, display, Mod5Mask, 0);
-  gok_keylabel_new (pKey, label, SPI_KEYMASK_MOD5, modmask);
-  /*  label = gok_key_label_from_keycode (keycode, display, 0, 1);
-  gok_keylabel_new (pKey, label, group1_mask, group1_mask);
-  label = gok_key_label_from_keycode (keycode, display, ShiftMask, 1);
-  gok_keylabel_new (pKey, label, "shift+group1");*/
-  g_free(label);
+  for (group = 0; group < 4; ++group) 
+  {
+	  for (level = 0; level < 4; ++level) 
+	  {
+		  /* this API is a little weird, gok_keylabel_new adds the label to the key */
+		  label = gok_key_label_from_keycode (keycode, display, level, group);
+		  if (!level && !group) /* bogus but we only have one group per key, not per label */
+		  {
+			  len = g_utf8_strlen (label, -1);
+			  if (len > 1) {
+				  if (len < 4) 
+					  pKey->FontSizeGroup = 3;
+				  else
+					  pKey->FontSizeGroup = FONT_SIZE_GROUP_UNIQUE;  
+			  }
+			  else {
+				  pKey->FontSizeGroup = FONT_SIZE_GROUP_GLYPH;
+			  }
+		  }
+		  gok_keylabel_new (pKey, label, level, group, NULL);
+	  }
+  }
+  /* XXX do we need to g_free (label)  ?? */
   snprintf (keycode_name, 8, "%d", keycode);
   pKey->pOutput = gok_output_new (OUTPUT_KEYCODE, keycode_name, SPI_KEY_PRESSRELEASE);
   return pKey;
@@ -532,7 +622,7 @@ GokKey* gok_key_new (GokKey* pKeyPrevious, GokKey* pKeyNext, GokKeyboard* pKeybo
 	pgok_key_new->pMimicKey = NULL;
 	pgok_key_new->pKeyNext = pKeyNext;
 	pgok_key_new->pKeyPrevious = pKeyPrevious;
-	pgok_key_new->pNodeAccessible = NULL;
+	pgok_key_new->accessible_node = NULL;
 	pgok_key_new->FontSizeGroup = FONT_SIZE_GROUP_UNDEFINED;
 	pgok_key_new->FontSize = -1; /* not set yet */
 	pgok_key_new->ComponentState.active = 0;
@@ -589,7 +679,7 @@ void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButto
 	{
 		gok_feedback_timer_stop_key_flash();
 	}
-	
+
 	/* unhook the key from the keyboard */
 	if ((pKeyboard != NULL) &&
 		(pKeyboard->pKeyFirst == pKey))
@@ -609,6 +699,8 @@ void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButto
 
 	/* delete all the key's labels */
 	pLabel = pKey->pLabel;
+	if (pLabel) gok_log ("deleting key %s %x (%x)\n", pLabel->Text ? pLabel->Text : "<empty>", pKey,
+			     (pKey->accessible_node) ? pKey->accessible_node->paccessible : 0xFFFF);
 	while (pLabel != NULL)
 	{
 		pLabelTemp = pLabel;
@@ -636,11 +728,19 @@ void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButto
 	gok_output_delete_all (pKey->pOutputWrapperPre);
 	gok_output_delete_all (pKey->pOutputWrapperPost);
 
+	if (pKey->accessible_node && pKey->accessible_node->paccessible) 
+	{
+		gok_log ("unreffing source");
+		gok_spy_accessible_unref (pKey->accessible_node->paccessible);
+	} 
+
 	g_free (pKey);
 
 	gok_feedback_drop_refs (pKey);
 	gok_scanner_drop_refs (pKey);
 	gok_repeat_drop_refs (pKey);
+
+	if (pKey->Type != KEYTYPE_WINDOW) g_free (pKey->pGeneral);
 }
 
 /**
@@ -652,7 +752,7 @@ void gok_key_delete (GokKey* pKey, GokKeyboard* pKeyboard, gboolean bDeleteButto
 **/ 
 gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 {
-	xmlChar* pStringAttributeValue;
+	xmlChar* pStringAttributeValue, *pStringSubAttributeValue;
 	xmlNode* pNodeKeyChild;
 	xmlNode* pNodeWrapperChild;
 	GokOutput* pNewOutput;
@@ -757,13 +857,13 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		pKey->Type = KEYTYPE_DOCK;
 		pKey->Style = KEYSTYLE_DOCK;
 		pKey->pGeneral = g_new0 (GokKeyboardDirection, 1);
-		pStringAttributeValue = 
+		pStringSubAttributeValue = 
 			xmlGetProp (pNode, (const xmlChar *) "dir");
-		if (pStringAttributeValue != NULL) 
+		if (pStringSubAttributeValue != NULL) 
 		{
 			* (GokKeyboardDirection *) pKey->pGeneral = 
-				gok_keyboard_parse_direction (pStringAttributeValue);
-			g_free (pStringAttributeValue);
+				gok_keyboard_parse_direction (pStringSubAttributeValue);
+			g_free (pStringSubAttributeValue);
 		}
 	}
 	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"move-resize") == 0)
@@ -771,13 +871,13 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		pKey->Type = KEYTYPE_MOVERESIZE;
 		pKey->Style = KEYSTYLE_NORMAL;
 		pKey->pGeneral = g_new0 (GokKeyboardDirection, 1);
-		pStringAttributeValue = 
+		pStringSubAttributeValue = 
 			xmlGetProp (pNode, (const xmlChar *) "dir");
-		if (pStringAttributeValue != NULL) 
+		if (pStringSubAttributeValue != NULL) 
 		{
 			* (GokKeyboardDirection *) pKey->pGeneral = 
-				gok_keyboard_parse_direction (pStringAttributeValue);
-			g_free (pStringAttributeValue);
+				gok_keyboard_parse_direction (pStringSubAttributeValue);
+			g_free (pStringSubAttributeValue);
 		}
 		pKey->is_repeatable = TRUE;
 	}
@@ -845,15 +945,17 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 			pKey->Style = KEYSTYLE_INSENSITIVE;
 		gok_mouse_control_init (pKey, pNode);
 	}
+	xmlFree (pStringAttributeValue);
 
 	/* branch target */
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "target");
 	{
 		if (pStringAttributeValue != NULL)
 		{
-			pKey->Target = (char *) pStringAttributeValue;
+			pKey->Target = (char *) g_strdup (pStringAttributeValue);
 		}
 	}
+	xmlFree (pStringAttributeValue);
 
 	/* font size group*/
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "fontsizegroup");
@@ -866,6 +968,7 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		else
 			pKey->FontSizeGroup = atoi ((char *)pStringAttributeValue);
 	}
+	xmlFree (pStringAttributeValue);
 
 	/* location of key (top, bottom, left & right) */
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "top");
@@ -878,6 +981,7 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		gok_log_x ("Can't find 'top' attribute in gok_key_initialize!\n");
 		return FALSE;
 	}
+	xmlFree (pStringAttributeValue);
 
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "bottom");
 	if (pStringAttributeValue != NULL)
@@ -889,6 +993,7 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		gok_log_x ("Warning: Can't find 'bottom' attribute in gok_key_initialize!\n");
 		return FALSE;
 	}
+	xmlFree (pStringAttributeValue);
 
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "left");
 	if (pStringAttributeValue != NULL)
@@ -900,6 +1005,7 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		gok_log_x ("Warning: Can't find 'left' attribute in gok_key_initialize!\n");
 		return FALSE;
 	}
+	xmlFree (pStringAttributeValue);
 
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "right");
 	if (pStringAttributeValue != NULL)
@@ -911,29 +1017,31 @@ gboolean gok_key_initialize (GokKey* pKey, xmlNode* pNode)
 		gok_log_x ("Warning: Can't find 'right' attribute in gok_key_initialize!\n");
 		return FALSE;
 	}
+	xmlFree (pStringAttributeValue);
 
 	/* is this a 'modifier' key? */
 	pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "modifier");
 	if (pStringAttributeValue != NULL)
 	{
 		/* add the modifier to the list of modifiers */
-		gok_modifier_add ((char *)pStringAttributeValue);
+		gok_modifier_add ((char *)g_strdup (pStringAttributeValue));
 		
 		/* store the name of the modifier on the key */
 		pKey->ModifierName = (gchar*)g_malloc (strlen ((char *)pStringAttributeValue) + 1);
 		strcpy (pKey->ModifierName, (char *)pStringAttributeValue);
 
 		/* what type of modifier is this? */
-		pStringAttributeValue = xmlGetProp (pNode, (const xmlChar *) "modifiertype");
-		if (pStringAttributeValue != NULL)
+		pStringSubAttributeValue = xmlGetProp (pNode, (const xmlChar *) "modifiertype");
+		if (pStringSubAttributeValue != NULL)
 		{
-	
-if (strcmp ((char *)pStringAttributeValue, "toggle") == 0)
-			{
-				gok_modifier_set_type (pKey->ModifierName, MODIFIER_TYPE_TOGGLE);
-			}
+		    if (strcmp ((char *)pStringSubAttributeValue, "toggle") == 0)
+		    {
+			gok_modifier_set_type (pKey->ModifierName, MODIFIER_TYPE_TOGGLE);
+		    }
+		    xmlFree (pStringSubAttributeValue);
 		}
 	}
+	xmlFree (pStringAttributeValue);
 
 	/* get child elements of the key */
 	pNodeKeyChild = pNode->xmlChildrenNode;
@@ -942,22 +1050,24 @@ if (strcmp ((char *)pStringAttributeValue, "toggle") == 0)
 		/* key label */
 		if (xmlStrcmp (pNodeKeyChild->name, (const xmlChar *)"label") == 0)
 		{
-			guint modifier = 0;
-			gchar *modifier_list = (gchar *) xmlGetProp (pNodeKeyChild, (const xmlChar *) "modifier");
-			gchar **tokens;
-			if (modifier_list) 
+			guint level = 0, group = 0;
+			gchar *levelname = (gchar *) xmlGetProp (pNodeKeyChild, (const xmlChar *) "level");
+			gchar *groupname = (gchar *) xmlGetProp (pNodeKeyChild, (const xmlChar *) "group");
+			gchar *vmodname = (gchar *) xmlGetProp (pNodeKeyChild, (const xmlChar *) "modifier");
+			gchar *content = xmlNodeGetContent (pNodeKeyChild);
+
+			if (levelname) 
 			{
-				gint i = 0;
-				tokens = g_strsplit (modifier_list, "+", 16);
-				while (tokens[i]) {
-					modifier |= gok_modifier_mask_for_name (gok_main_display (),
-										tokens[i]);
-					++i;
-				}
-				g_strfreev (tokens);
+			    level = atoi ((char *) levelname);
 			}
-			gok_keylabel_new (pKey, g_strstrip (xmlNodeGetContent (pNodeKeyChild)),
-					  modifier, modmask);
+			if (groupname) 
+			{
+			    group = atoi ((char *) groupname);
+			}
+			
+			gok_keylabel_new (pKey, g_strstrip (content),
+					  level, group, vmodname);
+			g_free (content);
 			pKey->has_text = TRUE;
 		}
 
@@ -1061,7 +1171,7 @@ if (strcmp ((char *)pStringAttributeValue, "toggle") == 0)
 *
 * returns: A pointer to the new key label, NULL if it wasn't created.
 **/
-GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, guint modifiers, guint mod_mask)
+GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, guint level, guint group, const gchar *vmods)
 {
 	GokKeyLabel* pNewLabel;
 	GokKeyLabel* pKeyLabel;
@@ -1070,14 +1180,15 @@ GokKeyLabel* gok_keylabel_new (GokKey* pKey, gchar* pLabelText, guint modifiers,
 	pNewLabel = (GokKeyLabel*) g_malloc(sizeof(GokKeyLabel));
 	
 	pNewLabel->Text = m_TextEmpty;
-	pNewLabel->modifiers = modifiers;
-	pNewLabel->mod_mask = mod_mask;
+	pNewLabel->level = level;
+	pNewLabel->group = group;
+	pNewLabel->vmods = g_strdup (vmods);
 	pNewLabel->pLabelNext = NULL;
 
 	if (pLabelText != NULL)
 	{
-		pNewLabel->Text = (gchar*)g_malloc (strlen(pLabelText) + 1);
-		strcpy (pNewLabel->Text, pLabelText);
+	        int len = strlen (pLabelText);
+		pNewLabel->Text = g_strdup (pLabelText);
 	}
 	
 	if (pKey != NULL)
@@ -1112,6 +1223,10 @@ void gok_keylabel_delete (GokKeyLabel* pKeyLabel)
 		{
 			g_free (pKeyLabel->Text);
 		}
+		if (pKeyLabel->vmods)
+		{
+			g_free (pKeyLabel->vmods);
+		}
 		g_free (pKeyLabel);
 	}
 }
@@ -1127,7 +1242,7 @@ void gok_keylabel_delete (GokKeyLabel* pKeyLabel)
 *
 * returns: TRUE if the key was initialized, FALSE if not.
 **/ 
-gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, guint modifiers, guint mod_mask)
+gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, guint level, guint group, const gchar *vmods)
 {
 	GokKeyLabel* pKeyLabel;
 		
@@ -1136,7 +1251,7 @@ gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, guint modifiers, gu
 	
 	if (pKey->pLabel == NULL)
 	{
-		gok_keylabel_new (pKey, pLabelText, modifiers, mod_mask);
+		gok_keylabel_new (pKey, pLabelText, level, group, vmods);
 	}
 	else
 	{
@@ -1144,7 +1259,7 @@ gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, guint modifiers, gu
 		pKeyLabel = pKey->pLabel;
 		while (pKeyLabel != NULL)
 		{
-			if ((pKeyLabel->modifiers == modifiers) && (pKeyLabel->mod_mask == mod_mask))
+			if ((pKeyLabel->level == level) && (pKeyLabel->group == group))
 			{
 				g_free (pKeyLabel->Text);
 				pKeyLabel->Text = (gchar*)g_malloc (strlen(pLabelText) + 1);
@@ -1157,7 +1272,7 @@ gboolean gok_key_add_label (GokKey* pKey, gchar* pLabelText, guint modifiers, gu
 		
 		if (pKeyLabel == NULL)
 		{
-			gok_keylabel_new (pKey, pLabelText, modifiers, mod_mask);
+			gok_keylabel_new (pKey, pLabelText, level, group, vmods);
 		}
 	}
 		
@@ -1176,7 +1291,7 @@ void gok_key_change_label (GokKey* pKey, gchar* LabelText)
 	g_assert (pKey != NULL);
 	g_assert (LabelText != NULL);
 	
-	gok_key_add_label (pKey, LabelText, 0, 0);
+	gok_key_add_label (pKey, LabelText, 0, 0, NULL);
 	
 	/* change the label text */
 	gok_key_set_button_label (pKey, LabelText);
@@ -1267,6 +1382,12 @@ void gok_key_update_label (GokKey* pKey)
 	/* if the new text is different from the old text then change it */
 	if (strcmp (pNewLabelText, pOldLabelText) != 0)
 	{
+#ifdef GOK_UTF8_DEBUG	    
+	    if (!g_utf8_validate (pNewLabelText, -1, NULL))
+		fprintf (stderr, "INVALID UTF8 : %s\n", pNewLabelText);
+	    fprintf (stderr, "setting button label: %s", pNewLabelText);
+	    gok_key_print_utf8_debug (pNewLabelText);
+#endif
 		gtk_label_set_text (GTK_LABEL(pButtonLabel), pNewLabelText);
 	}
 
@@ -1397,14 +1518,34 @@ gok_key_create_image_widget (GokKey* pKey)
 **/
 gchar* gok_key_get_label (GokKey* pKey)
 {
-	/* FIXME: make this work properly with multi-group keyboards! */
 	gchar* pTextReturned;
 	gchar* pToken;
 	gchar buffer[150];
 	GokKeyLabel* pKeyLabel;
 	gboolean bFoundLabel;
+	int  type, keycode;
+	guint level;
+	guint group = gok_key_get_effective_group ();
 	guint mods = gok_spy_get_modmask ();
-	
+	XkbDescRec *xkb;
+
+	xkb = gok_keyboard_get_xkb_desc (),
+	keycode = gok_output_get_keycode (pKey->pOutput);
+	type = gok_key_get_xkb_type_index (xkb, keycode, group);
+
+#ifdef GOK_DEBUG
+	char *typename = XGetAtomName (gok_keyboard_get_display (), xkb->map->types[type].name);
+	fprintf (stderr, "keycode %d, group %d, type=%d [%s]\n", keycode, group, type, typename);
+#endif
+	level = gok_key_level_for_type (gok_keyboard_get_display (), 
+					xkb,
+					type, &mods);
+
+#ifdef GOK_DEBUG
+	if (pKey->pLabel) fprintf (stderr, "Key %s: type %d, level %d [mods %x, preserved %x]\n", 
+				pKey->pLabel->Text,
+				type, level, gok_spy_get_modmask (), mods);
+#endif
 	g_assert (pKey != NULL);
 	
 	pTextReturned = m_TextEmpty;
@@ -1413,7 +1554,7 @@ gchar* gok_key_get_label (GokKey* pKey)
 	pKeyLabel = pKey->pLabel;
 	while (pKeyLabel != NULL)
 	{
-		if (pKeyLabel->modifiers == 0)
+		if ((pKeyLabel->level == 0) && (pKeyLabel->group == 0))
 		{
 			pTextReturned = pKeyLabel->Text;
 			break;
@@ -1429,26 +1570,15 @@ gchar* gok_key_get_label (GokKey* pKey)
 		return pTextReturned;
 	}
 	
-	/* find the correct label text for the current modifiers */
-	/* look through all the labels for one that best matches the modifiers */
+	/* find the correct label text for the current level and group */
 	bFoundLabel = FALSE;
 	pKeyLabel = pKey->pLabel;
 	while (pKeyLabel != NULL)
 	{
-		if ((pKeyLabel->modifiers) == (mods & pKeyLabel->mod_mask)) 
+		if (pKeyLabel->level == level && pKeyLabel->group == group) 
 		{
 			return pKeyLabel->Text;
 		}
-		else if ((bFoundLabel == FALSE) &&
-				(pKeyLabel->modifiers != 0))
-		{
-			/* the first modifier state that matches will be used */
-			if (pKeyLabel->modifiers & mods & pKeyLabel->mod_mask)
-			{
-				pTextReturned = pKeyLabel->Text;
-				bFoundLabel = TRUE;
-			}
-		}
 		pKeyLabel = pKeyLabel->pLabelNext;
 	}
 	
