@@ -839,9 +839,9 @@ void gok_feedbacks_update_sound_combo()
 **/
 void gok_page_feedbacks_get_sound_file ()
 {
-	GtkFileSelection* pDialogFilename;
+	GtkWidget* pDialogFilename;
 	gint response;
-	const gchar* filename;
+	gchar* filename;
 	GtkWidget* pEntry;
 
 	/* make sure the current feedback is not NULL */
@@ -851,21 +851,27 @@ void gok_page_feedbacks_get_sound_file ()
 	}
 	
 	/* create the file selector dialog */
-	pDialogFilename = (GtkFileSelection*)gtk_file_selection_new (_("Select sound file"));
+	pDialogFilename = gtk_file_chooser_dialog_new (_("Select sound file"),
+						       NULL,
+						       GTK_FILE_CHOOSER_ACTION_OPEN,
+						       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+						       GTK_STOCK_OK, GTK_RESPONSE_OK,
+						       NULL);
 
-   /* display that file selector dialog */
-	response = gtk_dialog_run (GTK_DIALOG(pDialogFilename));
+	/* display that file selector dialog */
+	gtk_window_set_default_size (GTK_WINDOW (pDialogFilename), 600, 400);
+	response = gtk_dialog_run (GTK_DIALOG (pDialogFilename));
 
-	if (response == GTK_RESPONSE_CANCEL)
+	if (response != GTK_RESPONSE_OK)
 	{
 		/* destroy the file selector dialog */
-		gtk_widget_destroy (GTK_WIDGET(pDialogFilename));
+		gtk_widget_destroy (pDialogFilename);
 
 		return;
 	}
 
 	/* get the file name */
-	filename = gtk_file_selection_get_filename (pDialogFilename);
+	filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (pDialogFilename));
 
 	/* add it to the feedback */
 	if (m_pFeedbackCurrent->pNameSound != NULL)
@@ -873,7 +879,7 @@ void gok_page_feedbacks_get_sound_file ()
 		g_free (m_pFeedbackCurrent->pNameSound);
 	}
         
-	m_pFeedbackCurrent->pNameSound = g_strdup (filename);
+	m_pFeedbackCurrent->pNameSound = filename;
 
 	/* update the combo box list */
 	gok_feedbacks_update_sound_combo();
@@ -886,7 +892,7 @@ void gok_page_feedbacks_get_sound_file ()
 	m_bIgnoreSoundNameChange = FALSE;
 
 	/* destroy the file selector dialog */
-	gtk_widget_destroy (GTK_WIDGET(pDialogFilename));
+	gtk_widget_destroy (pDialogFilename);
 }
 
 /**
