@@ -36,23 +36,35 @@ GOKSPY_FOCUS_EVENT,
 GOKSPY_DEFUNCT_EVENT,
 GOKSPY_WINDOW_ACTIVATE_EVENT,
 GOKSPY_WINDOW_DEACTIVATE_EVENT,
-GOKSPY_STATE_EVENT
+GOKSPY_STATE_EVENT,
+GOKSPY_KEYMAP_EVENT
 };
 
+typedef union
+{
+	guint value;
+	struct {
+		guint is_link:1;
+		guint is_ui:1;
+		guint is_menu:1;
+		guint has_context_menu:1;
+		guint is_toolbar_item:1;
+	} data;
+} AccessibleNodeFlags;
+
 /* this structure used in the creation of a linked list of Accessible*, char* (name) pairs */
 typedef struct AccessibleNode
 {
-    struct AccessibleNode* pnext;
     Accessible* paccessible;
-    guint is_link:1;
-    guint link:7; /* FIXME: need to generalize with type-safety, maybe a union? */
+    AccessibleNodeFlags flags;
+    guint link;
     char* pname;
 } AccessibleNode;
 
 /* this structure intended for use with a GQueue */
 typedef struct EventNode
 {
-	Accessible* paccessible;
+	const AccessibleEvent *event;
 	gint type; /* easier than dealing with string data */
 } EventNode;
 
