@@ -56,7 +56,7 @@ gboolean gok_windowlister_show()
 	GokKey* pKey;
 	GokKey* pKeyPrevious;
 	int column;
-	gchar *window_name = NULL;
+	const gchar *window_name = NULL;
     GList *window_list_entry;
 
 /*    GType window_type; */
@@ -149,8 +149,7 @@ gboolean gok_windowlister_show()
 			pKey->Left = column;
 			pKey->Right = column + 1;
 			window_name = wnck_window_get_name (window_list_entry->data);
-			if (!window_name) window_name = g_strdup ("");
-			gok_key_add_label (pKey, window_name, 0, 0, NULL);
+			gok_key_add_label (pKey, window_name ? g_strdup (window_name) : g_strdup (""), 0, 0, NULL);
 			
 			pKey->pGeneral = window_list_entry->data;  
 	
